package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.ConnectionException;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.transport.TransportException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 11:51
 */
public class ExecuteAnsibleServicesState extends SlicedState<ExecuteAnsibleServicesState.ExecuteAnsibleServiceStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExecuteAnsibleServicesState.class);

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;
    @Nonnull private final GeniUserProvider geniUserProvider;


    @Nonnull private final List<ExecuteAnsibleServiceStateSlice> slices;

    protected ExecuteAnsibleServicesState(@Nonnull SetupSoftwareExperimentJob job,
                                          @Nonnull GeniUserProvider geniUserProvider) {
        super("Execute ansible services");
        this.job = job;
        this.geniUserProvider = geniUserProvider;
        this.experiment = job.getExperiment();
        experiment.requireSlice();

        /////////////

        List<Pair<RspecNode, AnsibleService>> ansibleServices = getRelevantAnsibleServices(experiment);
        this.slices = ansibleServices.stream()
                .map(p -> new ExecuteAnsibleServiceStateSlice(job, p.getKey(), p.getValue()))
                .collect(Collectors.toList());
    }

    @Nonnull
    private static List<Pair<RspecNode, AnsibleService>> getRelevantAnsibleServices(Experiment experiment) {
        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        ModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec();
        assert model != null;

        if (model == null) {
            LOG.warn("ExecuteAnsibleServicesState will do nothing because there is no usable request RSpec");
            return Collections.emptyList();
        }

        List<Pair<RspecNode, AnsibleService>> res = new ArrayList<>();
        for (RspecNode node : model.getNodes()) {
            if (node.getAnsibleServices() != null) {
                res.addAll(
                        node.getAnsibleServices().stream()
                                .filter(Objects::nonNull)
                                .filter(as -> as.getInventory() != null ||
                                        as.getExecutePlaybook() != null)
                                .map(as -> new Pair<RspecNode, AnsibleService>(node, as))
                                .collect(Collectors.toList())
                );
            }
        }
        return res;
    }

    public static boolean hasAnsibleServices(Experiment experiment) {
        return !getRelevantAnsibleServices(experiment).isEmpty();
    }

    @Override
    public Collection<ExecuteAnsibleServiceStateSlice> getSlices() {
        return slices;
    }

    public class ExecuteAnsibleServiceStateSlice extends StateSlice {

        @Nonnull protected final RspecNode node;
        @Nonnull  protected final AnsibleService ansibleService;
        @Nonnull  protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;

        @Nullable  private SSHClient ssh = null;
        @Nullable private ExternalFileUtil externalFileUtil;

        public ExecuteAnsibleServiceStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                               @Nonnull RspecNode node,
                                               @Nonnull AnsibleService ansibleService) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.node = node;
            this.ansibleService = ansibleService;
        }

        @Override
        public String getName() {
            return "Execute ansible-playbook " + ansibleService.getExecutePlaybook() + " on " + node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            externalFileUtil = new ExternalFileUtil(ssh);

            if (ansibleService.getInventory() != null)
                setAndRunState(new UploadInventoryState());
            else
                LOG.debug("No inventory-path specified. Skipping upload inventory");

            setAndRunState(new ExecuteAnsibleServiceState());

            setupSoftwareExperimentJob.closeSSH(node);

            return null;
        }

        public class UploadInventoryState extends State {

            @Nonnull private final ModelRspec request;
            @Nonnull private final ModelRspec manifest;

            protected UploadInventoryState() {
                super("Uploading inventory on " + node.getClientId());

                //initialize request and manifest-model
                assert experiment.getNewRequestRspecSource() != null;
                ModelRspec request = experiment.getNewRequestRspecSource() == null ?
                        null :
                        experiment.getNewRequestRspecSource().getModelRspec();

                if (request == null) {
                    LOG.error("Cannot create inventory file because parsing request rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing request rspec failed");
                }

                this.request = request;


                assert experiment.getSliceOrNull() != null;
                ModelRspec manifest =
                        experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getManifestRspec() == null ?
                                null :
                                experiment.getSliceOrNull().getManifestRspec().getModelRspec();

                if (manifest == null) {
                    LOG.error("Cannot create inventory file because parsing manifest rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing manifest rspec failed");
                }
                this.manifest = manifest;

                if (experiment.getSliceOrNull().getManifestRspec().getStringRspec() == null){
                    LOG.error("Cannot create inventory file because parsing manifest rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing manifest rspec failed");
                }
            }

            public String createInventoryFile() {
                //construct a multimap containing the nodes grouped per ansible-group


                Multimap<String, RspecNode> nodesPerGroup = HashMultimap.create();
                for (RspecNode node : request.getNodes()) {
                    if (!node.getAnsibleGroups().isEmpty()) {
                        node.getAnsibleGroups().forEach(groupName -> nodesPerGroup.put(groupName, node));
                    } else {
                        nodesPerGroup.put(null, node);
                    }
                }

                //now write the inventory file
                StringBuilder inventory = new StringBuilder();

                //first write the groupless nodes
                if (nodesPerGroup.containsKey(null)) {
                    Collection<RspecNode> grouplessNodes = nodesPerGroup.removeAll(null);
                    grouplessNodes.forEach(node -> writeNodeToInventory(node, inventory));
                    inventory.append('\n');
                }

                //now write all groups

                for (String groupName : nodesPerGroup.keySet()) {
                    inventory.append("[").append(groupName).append("]\n");
                    nodesPerGroup.get(groupName).forEach(node -> writeNodeToInventory(node, inventory));
                    inventory.append('\n');
                }

                return inventory.toString();
            }

            private void writeNodeToInventory(RspecNode node, StringBuilder inventory) {
                final BasicStringRspec.LoginService loginService =
                        setupSoftwareExperimentJob.getBestNodeLoginFinder()
                                .findBestLogin(node.getUniqueId());

                if (loginService != null) {
                    inventory.append(node.getClientId())
                            .append("\tansible_host=").append(loginService.getHostname());

                    if (loginService.getPort() != 22) {
                        inventory.append("\tansible_port=").append(loginService.getPort());
                    }

                    if (!Objects.equals(loginService.getUsername(),
                            geniUserProvider.getLoggedInGeniUser().getUserUrn().getResourceName())) {
                        inventory.append("\tansible_user=").append(loginService.getUsername());
                    }
                    inventory.append('\n');
                }
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                String inventoryFilename;
                if (ansibleService.getInventory() != null) {
                    inventoryFilename = ansibleService.getInventory();
                } else {
                    inventoryFilename= "/etc/ansible/hosts"; //ansible default inventory file location
                }

                try {
                    //create inventory file
                    String inventory = createInventoryFile();

                    ssh.newSCPFileTransfer().upload(
                            new StringMemorySourceFile(externalFileUtil.getFilename(inventoryFilename), inventory),
                            externalFileUtil.getFolder(inventoryFilename));

                } catch (IOException ex) {
                    LOG.error("Error while uploading inventory-file to {} on {}", inventoryFilename, node.getClientId(), ex);
                    updateMessage(String.format("Error while uploading inventory-file to %s on %s: %s", inventoryFilename, node.getClientId(), ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class ExecuteAnsibleServiceState extends State {

            private final StringProperty galaxyInputStream, galaxyErrorStream;
            private final StringProperty playbookInputStream, playbookErrorStream;


            protected ExecuteAnsibleServiceState() {
                super("Executing ansible-playbook '" + ansibleService.getExecutePlaybook() + "' on " + node.getClientId());

                this.galaxyInputStream = new SimpleStringProperty("");
                this.galaxyErrorStream = new SimpleStringProperty("");
                this.playbookInputStream = new SimpleStringProperty("");
                this.playbookErrorStream = new SimpleStringProperty("");

            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {

                try {
                    //first install the requirements if needed
                    if (ansibleService.getInstallRequirements() != null) {
                        int dividerPosition = ansibleService.getInstallRequirements().lastIndexOf('/');
                        String folder = ansibleService.getInstallRequirements().substring(0, dividerPosition);
                        String filename = ansibleService.getInstallRequirements().substring(dividerPosition + 1);

                        String galaxyCommand = ansibleService.getGalaxyCommand();
                        if (galaxyCommand == null)
                            galaxyCommand = "ansible-galaxy";

                        String galaxyCommandString = String.format("cd %s && %s install -r %s -p roles",
                                folder, galaxyCommand, filename);
                        LOG.debug("Executing galaxy command: "+galaxyCommandString);
                        Session.Command c =
                                ssh.startSession().exec(galaxyCommandString);

                        new InputStreamToStringPropertyReaderThread(c.getInputStream(), galaxyInputStream, Level.INFO, "GALAXY-OUTPUT: ").start();
                        new InputStreamToStringPropertyReaderThread(c.getErrorStream(), galaxyErrorStream, Level.WARN, "GALAXY-ERROR: ").start();

                        c.join();

                        if (c.getExitStatus() != 0) {
                            updateMessage("ansible-galaxy execution was unsuccessful. Aborting");
                            return ExperimentTaskStatus.FAILED;
                        }
                    }

                    assert ansibleService.getExecutePlaybook() != null;
                    int dividerPosition = ansibleService.getExecutePlaybook().lastIndexOf('/');
                    String folder = ansibleService.getExecutePlaybook().substring(0, dividerPosition);
                    String filename = ansibleService.getExecutePlaybook().substring(dividerPosition + 1);

                    String playbookCommand = ansibleService.getPlaybookCommand();
                    if (playbookCommand == null)
                        playbookCommand = "ansible-playbook";

                    String arguments = "";
                    arguments += filename;

                    if (Objects.equals(ansibleService.getDebug(), Boolean.TRUE))
                        arguments += " -v";

                    if (ansibleService.getInventory() != null)
                        arguments += " -i '"+ansibleService.getInventory()+"'";

                    //this must be the LAST argument, as it is a redirect and pipe
                    if (ansibleService.getLogfile() != null)
                        arguments += " 2>&1 | tee '"+ansibleService.getLogfile()+"'";

                    String playbookCommandString = String.format("cd %s && %s %s", folder, playbookCommand, arguments);
                    LOG.debug("Executing playbook command: "+playbookCommandString);
                    Session.Command c =
                            ssh.startSession().exec(playbookCommandString);

                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), playbookInputStream, Level.INFO, "PLAYBOOK-OUTPUT: ").start();
                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), playbookErrorStream, Level.WARN, "PLAYBOOK-ERROR: ").start();

                    c.join();


                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-playbook execution was unsuccessful.");
                        return ExperimentTaskStatus.FAILED;
                    }


                } catch (TransportException | ConnectionException ex) {
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            public StringProperty getGalaxyInputStream() {
                return galaxyInputStream;
            }

            public StringProperty getGalaxyErrorStream() {
                return galaxyErrorStream;
            }

            public StringProperty getPlaybookInputStream() {
                return playbookInputStream;
            }

            public StringProperty getPlaybookErrorStream() {
                return playbookErrorStream;
            }

            public RspecNode getNode() {
                return node;
            }

            public AnsibleService getAnsibleService() {
                return ansibleService;
            }
        }
    }

    private static class InputStreamToStringPropertyReaderThread extends Thread {

        private final InputStream inputStream;
        private final StringProperty output;

        private final String logPrefix;
        private final LogFunction logFunction;


        private InputStreamToStringPropertyReaderThread(InputStream inputStream, StringProperty output,
                                                        @Nullable Level logLevel, String logPrefix) {
            this.inputStream = inputStream;
            this.output = output;
            this.logPrefix = logPrefix;

            if (logLevel != null) {
                switch (logLevel) {
                    case TRACE:
                        logFunction = LOG::trace;
                        break;
                    case DEBUG:
                        logFunction = LOG::debug;
                        break;
                    case INFO:
                        logFunction = LOG::info;
                        break;
                    case WARN:
                        logFunction = LOG::warn;
                        break;
                    case ERROR:
                        logFunction = LOG::error;
                        break;
                    default:
                        logFunction = null;
                }

            } else
                logFunction = null;
        }

        @Override
        public void run() {

            try {
                BufferedReader stdInput = new BufferedReader(new InputStreamReader(inputStream));

                String s;
                while ((s = stdInput.readLine()) != null) {
                    System.out.println(s);
                    String str = s;

                    if (logFunction != null) {
                        logFunction.log(logPrefix + str);
                    }

                    Platform.runLater(() -> output.setValue(output.get() + "\n" + str));

                }

                inputStream.close();
            } catch (IOException ex) {
                LOG.error("Error while reading from InputStream to StringProperty", ex);
            }

        }

        private interface LogFunction {
            void log(String text);
        }
    }

}
