package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.channel.direct.Session;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 11:51
 */
public class WaitForExecuteFinishedState extends SlicedState<WaitForExecuteFinishedState.WaitForExecuteServiceStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(WaitForExecuteFinishedState.class);
    public static final int MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES = 60;
    public static final int TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS = 10_000;


    @Nonnull
    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    @Nonnull
    private final Experiment experiment;


    @Nonnull
    private final List<WaitForExecuteServiceStateSlice> slices;


    protected WaitForExecuteFinishedState(@Nonnull SetupSoftwareExperimentJob job) {
        super("Execute ansible services");
        this.setupSoftwareExperimentJob = job;
        this.experiment = job.getExperiment();
        experiment.requireSlice();

        List<Pair<RspecNode, ExecuteService>> executeServices =
                getExecuteServicesToWaitFor(experiment);

        slices = executeServices.stream()
                        .map(p -> new WaitForExecuteServiceStateSlice(job, p.getKey(), p.getValue()))
                        .collect(Collectors.toList());
    }

    @Nonnull
    private static List<Pair<RspecNode, ExecuteService>> getExecuteServicesToWaitFor(Experiment experiment) {
        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        ModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec();
        assert model != null;

        if (model == null) {
            LOG.warn("ExecuteAnsibleServicesState will do nothing because there is no usable request RSpec");
            return Collections.emptyList();
        }

        List<Pair<RspecNode, ExecuteService>> res = new ArrayList<>();
        for (RspecNode node : model.getNodes()) {
            if (node.getExecuteServices() != null) {
                res.addAll(
                        node.getExecuteServices().stream()
                                .filter(Objects::nonNull)
                                .filter(es -> es.getFinishedFile() != null)
                                .map(es -> new Pair<RspecNode, ExecuteService>(node, es))
                                .collect(Collectors.toList())
                );
            }
        }
        return res;
    }

    public static boolean hasWaitForExecute(Experiment experiment) {
        return !getExecuteServicesToWaitFor(experiment).isEmpty();
    }

    @Override
    public Collection<WaitForExecuteServiceStateSlice> getSlices() {
        return slices;
    }

    public class WaitForExecuteServiceStateSlice extends StateSlice {

        @Nonnull
        protected final RspecNode node;
        @Nonnull
        protected final ExecuteService executeService;

        @Nullable
        private SSHClient ssh = null;
        @Nullable
        private ExternalFileUtil externalFileUtil;

        public WaitForExecuteServiceStateSlice(@Nonnull Job<?> job,
                                               @Nonnull RspecNode node,
                                               @Nonnull ExecuteService executeService) {
            super(job);
            this.node = node;
            this.executeService = executeService;
        }

        @Override
        public String getName() {
            return "Waiting for ExecuteServices to finish on " + node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            externalFileUtil = new ExternalFileUtil(ssh);

            setAndRunState(new WaitForExecuteServicesState());

            setupSoftwareExperimentJob.closeSSH(node);

            return null;
        }

        public class WaitForExecuteServicesState extends State {

            protected WaitForExecuteServicesState() {
                super("Waiting for execute-services to finish on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                try {
                    if (executeService.getFinishedFile() == null) {
                        LOG.warn("BUG: executeService.getFinishedFile() == null so nothing to do.");
                        return ExperimentTaskStatus.SUCCESS;
                    }

                    boolean fileFound = false;
                    int tries = 0;

                    while (!fileFound && tries < MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES) {
                        Session.Command command =
                                ssh.startSession().exec("test -f " + executeService.getFinishedFile());
                        command.join();

                        fileFound = command.getExitStatus() == 0;

                        tries++;
                        if (!fileFound) {
                            updateMessage("Looked for '" + executeService.getFinishedFile() + "' " + tries + " times.");
                            Thread.sleep(TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS);
                        }
                    }

                    if (!fileFound) {
                        updateMessage("Timed out while waiting for finished-file '" + executeService.getFinishedFile() + "' to appear");
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        updateMessage("Found finished-file '" + executeService.getFinishedFile() + "'");
                    }


                    updateMessage("Found finished-file "+executeService.getFinishedFile());
                    return ExperimentTaskStatus.SUCCESS;
                } catch (IOException ex) {
                    LOG.error("Error while waiting for finished-files on {}", node.getClientId(), ex);
                    updateMessage("Error while waiting for finished-files on {}" + ex.getMessage());
                    return ExperimentTaskStatus.FAILED;
                }
            }
        }
    }
}
